/* cdLookup.C

   Copyright (C) 1999 by David Rose.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   To contact the original author of this program, write to:
           David Rose
           224 W. Dryden St Apt. 418
	   Glendale, CA 91202
	   USA

	   drose@lafn.org
   */

#include "system.h"

#include "cdLookup.h"
#include "cdDialog.h"
#include "database.h"
#include "dataCell.h"
#include "artist.h"
#include "format.h"
#include "playlistDialog.h"
#include "infoMessage.h"

static const int num_clist_titles = 2;
static const gchar *
clist_titles[num_clist_titles] = { "CD Title", "Performer" };

CDLookup::AcceptDialogData::
AcceptDialogData(CDDialog *dialog, int cd_index) :
  _dialog(dialog), _cd_index(cd_index) {
}

CDLookup::
CDLookup(const Dataset &data, bool edit_mode) :
  PalomaDialog(data),
  _edit_mode(edit_mode),
  _menu("<cd-lookup>"),
  _clist(num_clist_titles, clist_titles),
  _ok_button("OK"),
  _edit_button("Edit"),
  _cancel_button("Cancel")
{
  layout_window();
  setup();
}

CDLookup::
~CDLookup() {
}

// get_cd() returns the currently selected CD, if any.
TableDataProxy CDLookup::
get_cd() const {
  if (_clist.selbegin() != _clist.selend()) {
    int row = *(_clist.selbegin());
    return TableDataProxy(TI_cds, _cd_list[row]);
  }

  return TableDataProxy(TI_cds);
}

const CDLookup::CDList &CDLookup::
get_cd_list() const {
  return _cd_list;
}

void CDLookup::
popup_cd_dialog(int cd_index) {
  assert(cd_index >= 0 && cd_index < _cd_list.size());
  int cd_id = _cd_list[cd_index];

  Dataset new_data = _data;
  new_data[TI_cds] = TableDataProxy(TI_cds, cd_id);
  new_data[TI_cds_extension] = 
    TableDataProxy(TI_cds_extension, cd_id);

  CDDialog *dialog = CDDialog::make_dialog(new_data, -1);
  connect_to_method(dialog->accepted, this, &cd_dialog_accepted, 
		    AcceptDialogData(dialog, cd_index));
}

void CDLookup::
menu_make_playlist() {
  // First, build up the list of CD id's we have selected.
  string cd_ids;
  for (Gtk_CList::seliterator si = _clist.selbegin();
       si != _clist.selend();
       ++si) {
    assert(*si >= 0 && *si < _cd_list.size());
    cd_ids += ", " + format_int(_cd_list[*si]);
  }

  if (cd_ids.empty()) {
    new InfoMessage
      ("You have no rows selected.  Use the mouse to select the CD's\n"
       "you want to put in your playlist, and choose this option again.");
  } else {
    // Now issue a cheesy select to get all those CD's.
    PlaylistDialog::make_dialog
      (_data, -1, 
       "SELECT cds.cd_id, cds.title"
       " FROM cds"
       " WHERE cds.cd_id IN (" + cd_ids.substr(2) + ")");
  }
}  

void CDLookup::
ok_button_clicked() {
  // In edit mode, the ok button brings up the selected CD if we have
  // exactly one row selected, and it brings up a playlist if we have
  // multiple rows selected.
  if (_edit_mode) {
    if (_clist.selbegin() != _clist.selend()) {
      Gtk_CList::seliterator si = _clist.selbegin();
      ++si;
      if (si == _clist.selend()) {
	// One row selected.
	int row = *(_clist.selbegin());
	popup_cd_dialog(row);
      } else {
	// Multiple rows selected.
	menu_make_playlist();
      }
    }
  }

  accepted();
  delete_self();
}

void CDLookup::
edit_button_clicked() {
  // Clicking the edit button means to edit the currently selected
  // row, if any.
  if (_clist.selbegin() != _clist.selend()) {
    int row = *(_clist.selbegin());
    popup_cd_dialog(row);
  }
}

void CDLookup::
cancel_button_clicked() {
  delete_self();
}

gint CDLookup::
clist_clicked(GdkEventButton *event) {
  // In edit mode, double-clicking a row means to pop up an edit
  // dialog for that row.  In non-edit mode, double-clicking a row
  // means to select the row and close the dialog.
  if (event->button == 1 && event->type == GDK_2BUTTON_PRESS) {
    int row, column;
    _clist.get_selection_info(event->x, event->y, &row, &column);
    if (row >= 0 && row < _cd_list.size()) {
      if (_edit_mode) {
	popup_cd_dialog(row);
      } else {
	_clist.cause_select_row(row, 0);
	accepted();
	delete_self();
      }
    }
  }

  return false;
}

void CDLookup::
cd_dialog_accepted(AcceptDialogData add) {
  const TableDataProxy &cd = add._dialog->get_data()[TI_cds];

  int t = add._cd_index;
  _clist.set_text(t, 0, cd[CFI_title].get_string());
  _clist.set_text
    (t, 1, get_formatted_artist_name(cd[CFI_performer].get_foreign(TI_artists)));
}

void CDLookup::
layout_window() {
  set_default_size(450, 450);
  set_title("CD Look-Up");

  // The menu bar.
  _menubar = _menu.get_menubar_widget("");
  assert(_menubar.get_object() != NULL);

  _menu.create_item("/File/Make a playlist", "", "",
		    ItemFactoryConnector<CDLookup, int>(this, &menu_make_playlist));
  _menu.create_item("/File/Edit", "", "",
		    ItemFactoryConnector<CDLookup, int>(this, &edit_button_clicked));
  _menu.create_item("/File/Accept", "", "",
		    ItemFactoryConnector<CDLookup, int>(this, &ok_button_clicked));
  _menu.create_item("/File/Close", "", "",
		    ItemFactoryConnector<CDLookup, int>(this, &cancel_button_clicked));

  vbox_.pack_start(*_menubar.get_object(), false, false);
  _menubar.get_object()->show();

  // The box for the main set of widgets.
  vbox_.pack_start(_box1);
  _box1.show();
  _box1.set_border_width(8);

  // The list of CDs.
  _cd_window.set_policy(GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  _box1.pack_start(_cd_window, true, true, 0);
  _cd_window.show();

  _cd_window.add_with_viewport(_clist);

  _clist.column_titles_passive();
  _clist.set_selection_mode(GTK_SELECTION_MULTIPLE);
  _clist.set_column_width(0, 240);
  _clist.set_column_width(1, 120);
  _clist.show();

  const gchar *row_text[num_clist_titles] = {
    NULL, NULL,
  };

  QueryResult qr = music_db->sql_command
    ("SELECT cd_id, title, performer "
     "FROM cds "
     "ORDER BY cds.title");

  int num_rows = qr.get_num_rows();
  for (int t = 0; t < num_rows; t++) {
    int cd_id = qr.get_cell(t, 0).get_int();
    string cd_title = qr.get_cell(t, 1).get_string();
    TableDataProxy artist = qr.get_cell(t, 2).get_foreign(TI_artists);

    string artist_name = get_formatted_artist_name(artist);

    row_text[0] = cd_title.c_str();
    row_text[1] = artist_name.c_str();

    _clist.append(row_text);
    _cd_list.push_back(cd_id);
  }

  connect_to_method(_clist.button_press_event, this,
		    &clist_clicked);

  // The OK button.
  _ok_button.set_usize(80, 36);
  action_area_.pack_start(_ok_button, true, false, 0);
  _ok_button.set_flags(GTK_CAN_DEFAULT);
  _ok_button.grab_default();
  _ok_button.show();
  connect_to_method(_ok_button.clicked, this, &ok_button_clicked);
  
  // The Edit button.
  _edit_button.set_usize(80, 36);
  action_area_.pack_start(_edit_button, true, false, 0);
  _edit_button.set_flags(GTK_CAN_DEFAULT);
  _edit_button.show();
  connect_to_method(_edit_button.clicked, this, &edit_button_clicked);
  
  // The Cancel button.
  _cancel_button.set_usize(80, 36);
  action_area_.pack_start(_cancel_button, true, false, 0);
  _cancel_button.set_flags(GTK_CAN_DEFAULT);
  _cancel_button.show();
  connect_to_method(_cancel_button.clicked, this, &cancel_button_clicked);
}
