/* importDialog.C

   Copyright (C) 1999 by David Rose.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   To contact the original author of this program, write to:
           David Rose
           224 W. Dryden St Apt. 418
	   Glendale, CA 91202
	   USA

	   drose@lafn.org
   */

#include "system.h"

// This needs to come first for namespace collision reasons.
#include <regex.h>

#include "importDialog.h"
#include "importListDialog.h"
#include "importList.h"
#include "queryResult.h"
#include "database.h"
#include "infoMessage.h"
#include "string_utils.h"
#include "guess_file_type.h"
#include "get_basename.h"

#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

static void
scan_tree(const string &root_prefix, const string &fullpath,
	  const string &relpath, const string &rel_prefix, int file_type,
	  bool use_re, regex_t *re, const string &separators,
	  ImportList *good_files, ImportList *bad_files);

ImportDialog::
ImportDialog(const Dataset &data) :
  PalomaDialog(data),
  _volume_label("Import into volume:"),
  _leave_button(NULL, "Leave files in place (root dir, below, must be within volume hierarchy)"),
  _copy_button(_leave_button, "Copy files into volume"),
  _convert_button(_copy_button, "Copy files and convert to: "),
  _root_label("Root directory:"),
  _delimiters_button(NULL, "Dividing characters"),
  _regexp_button(_delimiters_button, "Regexp"),
  _type_label("Source files are of type: "),
  _ok_button("Get files"),
  _cancel_button("Cancel")
{
  // Get all of the soundfile definitions, except the CD-Rom definition.
  QueryResult qr = music_db->sql_command
    ("SELECT soundfile_id, description"
     " FROM soundfile, configure"
     " WHERE soundfile_id != configure.cdrom_type"
     " ORDER BY description");
  int num_rows = qr.get_num_rows();
  int r;
  _type_pick.add_choice(-1, "Guess");
  for (r = 0; r < num_rows; r++) {
    int soundfile_id = qr.get_cell(r, 0).get_int();
    string description = qr.get_cell(r, 1).get_string();
    _convert_type.add_choice(soundfile_id, description);
    _type_pick.add_choice(soundfile_id, description);
  }
  _convert_type.finish_choices();
  _type_pick.finish_choices();

  // Get all of the available volumes.
  qr = music_db->sql_command
    ("SELECT volume_id, description FROM volumes"
     " ORDER BY description");
  num_rows = qr.get_num_rows();
  for (r = 0; r < num_rows; r++) {
    int volume_id = qr.get_cell(r, 0).get_int();
    string description = qr.get_cell(r, 1).get_string();
    _volume_pick.add_choice(volume_id, description);
  }
  _volume_pick.finish_choices();

  _filename_entry.set_text("/.");

  layout_window();
  setup();
}

ImportDialog::
~ImportDialog() {
}

void ImportDialog::
ok_button_clicked() {
  Volume volume(_data, _volume_pick.get_item(-1));
  ImportListDialog::Disposition disp = ImportListDialog::D_leave;
  int convert_to_type = -1;
  int source_type = _type_pick.get_item(-1);

  if (_copy_button.get_active()) {
    disp = ImportListDialog::D_copy;
  } else if (_convert_button.get_active()) {
    disp = ImportListDialog::D_convert;
    convert_to_type = _convert_type.get_item(-1);
    if (convert_to_type < 0) {
      new InfoMessage("You must choose a type to convert to.");
      return;
    }
  }

  if (!volume.is_valid()) {
    new InfoMessage("You must define a volume to store the files in.");
    return;
  }

  if (!volume.exists()) {
    new InfoMessage("Volume " + volume.get_description() +
		    " does not exist.");
    return;
  }

  string root_dir = _root_entry.get_filename();
  string root_prefix;
  string start_file;
  struct stat statbuf;
  if (stat(root_dir.c_str(), &statbuf) < 0) {
    new InfoMessage("Root directory " + root_dir + " does not exist.");
    return;
  }
    
  if (S_ISDIR(statbuf.st_mode)) {
    if (access(root_dir.c_str(), R_OK | X_OK) != 0) {
      new InfoMessage("Cannot read directory " + root_dir);
      return;
    }
    root_prefix = root_dir + "/";
    start_file = "";

  } else if (S_ISREG(statbuf.st_mode)) {
    if (access(root_dir.c_str(), R_OK) != 0) {
      new InfoMessage("Cannot read file " + root_dir);
      return;
    }
    // We've named a particular file instead of a directory.  Extract
    // out just the filename portion.
    root_prefix = get_dirname(root_dir);
    start_file = get_basename(root_dir);

  } else {
    new InfoMessage("Not a directory or file: " + root_dir);
    return;
  }

  if (disp == ImportListDialog::D_leave) {
    string volume_def = volume.get_definition();
    if (root_prefix.substr(0, volume_def.length()) != volume_def) {
      new InfoMessage(root_prefix + " is not within volume " +
		      volume.get_description());
      return;
    }
		      
  } else {
    if (!volume.can_store_files()) {
      new InfoMessage("Cannot store files in " + volume.get_description());
      return;
    }
  }

  bool use_re;
  string separators;
  regex_t re;

  if (_delimiters_button.get_active()) {
    use_re = false;
    separators = _filename_entry.get_text();

  } else {
    use_re = true;
    string exp_string = trim_blanks(_filename_entry.get_text());
    int result = regcomp(&re, exp_string.c_str(), 0);
    if (result != 0) {
      // Error compiling regular expression.
      char error_msg[256] = "Invalid regular expression.";
      regerror(result, &re, error_msg, 256);
      new InfoMessage(error_msg);
      return;
    }
  }

  // Now scan the file hierarchy beginning at root_dir.
  ImportList *good_files = new ImportList;
  ImportList *bad_files = new ImportList;
  good_files->_columns.push_back(string("1"));
  bad_files->_columns.push_back(string("Filename"));
  scan_tree(root_prefix, root_prefix + start_file, 
	    start_file, "", source_type,
	    use_re, &re, separators,
	    good_files, bad_files);

  if (use_re) {
    regfree(&re);
  }

  new ImportListDialog(_data, this, volume, disp, convert_to_type,
		       root_prefix, good_files, bad_files);
}

void ImportDialog::
cancel_button_clicked() {
  delete_self();
}

void ImportDialog::
layout_window() {
  set_title("Import song files");
  set_default_size(450, 0);

  // The box for the main set of widgets.
  vbox_.pack_start(_box1);
  _box1.show();
  _box1.set_border_width(8);

  // Volume, label and entry.
  _volume_label.set_alignment(0, 0);
  _box1.pack_start(_volume_label, false, false, 0);
  _box1.pack_start(_volume_pick);
  _volume_label.show();
  _volume_pick.show();

  // The type radio buttons.
  _box1.pack_start(_rframe, true, true, 8);
  _rframe.show();
  _rframe.set_shadow_type(GTK_SHADOW_ETCHED_IN);
  _rframe.add(_rbox);
  _rbox.show();
  _rbox.set_border_width(8);
  _rbox.pack_start(_leave_button);
  _leave_button.show();
  _rbox.pack_start(_copy_button);
  _copy_button.show();
  _rbox.pack_start(_cbox);
  _cbox.show();
  _cbox.pack_start(_convert_button, false, false);
  _convert_button.show();
  _cbox.pack_start(_convert_type);
  _convert_type.show();

  // Root directory, label and entry.
  _root_label.set_alignment(0, 0);
  _box1.pack_start(_root_label, false, false, 0);
  _box1.pack_start(_root_entry);
  _root_label.show();
  _root_entry.show();

  // Filename pattern, label and entry.
  _box1.pack_start(_fbox);
  _fbox.pack_start(_delimiters_button, false);
  _fbox.pack_start(_regexp_button, false);
  _box1.pack_start(_filename_entry);
  _fbox.show();
  _delimiters_button.show();
  _regexp_button.show();
  _filename_entry.show();

  // The source file type.
  _box1.pack_start(_cbox2, true, true, 2);
  _cbox2.show();
  _cbox2.pack_start(_type_label, false, false);
  _cbox2.pack_start(_type_pick);
  _type_label.show();
  _type_pick.show();

  // The OK button.
  _ok_button.set_usize(80, 36);
  action_area_.pack_start(_ok_button, true, false, 0);
  _ok_button.set_flags(GTK_CAN_DEFAULT);
  _ok_button.grab_default();
  _ok_button.show();
  connect_to_method(_ok_button.clicked, this, &ok_button_clicked);
  
  // The Cancel button.
  _cancel_button.set_usize(80, 36);
  action_area_.pack_start(_cancel_button, true, false, 0);
  _cancel_button.set_flags(GTK_CAN_DEFAULT);
  _cancel_button.show();
  connect_to_method(_cancel_button.clicked, this, &cancel_button_clicked);
}

static void
separate_filename(const string &relpath, int type, 
		  const string &separators, ImportList *good_files) {
  good_files->_rows.push_back(ImportList::Row(relpath, type));
  ImportList::Row &row = good_files->_rows.back();

  int last, next;
  last = 0;
  next = relpath.find_first_of(separators);
  while (next != string::npos) {
    row._data.push_back(relpath.substr(last, next - last));
    last = next + 1;
    next = relpath.find_first_of(separators, last);
  }
  row._data.push_back(relpath.substr(last));
  good_files->check_min_columns(row._data.size());
}

static void
match_re(const string &relpath, int type, regex_t *re,
	 ImportList *good_files, ImportList *bad_files) {
  // A regular file.
  vector<regmatch_t> matches(re->re_nsub + 1);
  int result = 
    regexec(re, relpath.c_str(), re->re_nsub + 1, &matches[0], 0);

  if (result == REG_NOMATCH) {
    // Failed to match.
    bad_files->_rows.push_back(ImportList::Row(relpath, type));
    ImportList::Row &row = bad_files->_rows.back();
    row._data.push_back(relpath);

  } else if (result == 0) {
    // Matched!
    good_files->_rows.push_back(ImportList::Row(relpath, type));
    ImportList::Row &row = good_files->_rows.back();

    if (re->re_nsub == 0) {
      // No bracketed subexpressions; use the matched string.
      row._data.push_back
	(relpath.substr(matches[0].rm_so,
			matches[0].rm_eo - matches[0].rm_so));
    } else {
      // One column for each subexpression.
      for (int i = 0; i < re->re_nsub; i++) {
	row._data.push_back
	  (relpath.substr(matches[i+1].rm_so,
			  matches[i+1].rm_eo - matches[i+1].rm_so));
      }
    }
    good_files->check_min_columns(row._data.size());

  } else {
    char error_msg[256] = "Error matching regular expression.";
    regerror(result, re, error_msg, 256);
    cerr << error_msg << "\n";
  }
}

static void
scan_tree(const string &root_prefix, const string &fullpath,
	  const string &relpath, const string &rel_prefix, int file_type,
	  bool use_re, regex_t *re, const string &separators,
	  ImportList *good_files, ImportList *bad_files) {
  struct stat statbuf;
  if (stat(fullpath.c_str(), &statbuf) < 0) {
    perror(fullpath.c_str());
    
  } else {
    if (S_ISDIR(statbuf.st_mode)) {
      // A nested subdirectory, recurse.
      DIR *subdp = opendir(fullpath.c_str());
      if (subdp == NULL) {
	cerr << "Couldn't read directory " << fullpath << "\n";
      } else {
	// Read the directory.
	struct dirent *dirp;
	dirp = readdir(subdp);
	while (dirp != NULL) {
	  if (dirp->d_name[0] != '.') {
	    string sub_relpath = rel_prefix + dirp->d_name;
	    string sub_fullpath = root_prefix + sub_relpath;
	    scan_tree(root_prefix, sub_fullpath,
		      sub_relpath, sub_relpath + "/", file_type,
		      use_re, re, separators,
		      good_files, bad_files);
	  }
	  
	  dirp = readdir(subdp);
	}
	closedir(subdp);
      }
      
    } else if (S_ISREG(statbuf.st_mode)) {
      // A regular file.
      int type = file_type;
      if (type < 0) {
	type = guess_file_type(fullpath);
      }
      if (type < 0) {
	cerr << "Couldn't guess type of " << fullpath << "\n";
	bad_files->_rows.push_back(ImportList::Row(relpath, -1));
	ImportList::Row &row = bad_files->_rows.back();
	row._data.push_back(relpath);
	
      } else {
	if (use_re) {
	  match_re(relpath, type, re, good_files, bad_files);
	} else {
	  separate_filename(relpath, type, separators, good_files);
	}
      }
    }
  }
}
