/* importingDialog.C

   Copyright (C) 1999 by David Rose.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   To contact the original author of this program, write to:
           David Rose
           224 W. Dryden St Apt. 418
	   Glendale, CA 91202
	   USA

	   drose@lafn.org
   */

#include "system.h"

#include "importingDialog.h"
#include "protect_from_shell.h"
#include "infoMessage.h"
#include "dataCell.h"
#include "queryResult.h"
#include "database.h"
#include "expand_variables.h"
#include "format.h"
#include "song.h"
#include "get_basename.h"

ImportingDialog::
ImportingDialog(const Dataset &data, const Volume &volume,
		int convert_to_type, const string &root_prefix,
		Songs &songs) :
  PalomaDialog(data),
  _volume(volume),
  _convert_to_type(convert_to_type),
  _root_prefix(root_prefix),
  _cancel_button("Cancel")
{
  // Steal all the songs from the list.
  _songs.swap(songs);

  _canceled = false;

  _progress_bar.configure(0.0, 0.0, _songs.size());
  _progress_bar.set_show_text(true);
  _current_index = 0;

  layout_window();
  setup();

  connect_to_method(_shell.success, this, &import_success);
  connect_to_method(_shell.failure, this, &import_failure);

  if (!import_current()) {
    // We'll set a timeout for a quarter second longer just to give
    // the window a chance to pop up before we destroy it.  (Surely
    // there's a better way!)
    connect_to_method(Gtk_Main::timeout(250), this, &delete_timeout);
  }
}

ImportingDialog::
~ImportingDialog() {
  Songs::iterator si;
  for (si = _songs.begin(); si != _songs.end(); ++si) {
    delete (*si);
  }
}

void ImportingDialog::
import_success() {
  string new_relpath = 
    _volume.close_write_file(_filename_to, _basename_to);
  time_t now = time(NULL);

  Song &import_song = *_songs[_current_index];
  TableDataProxy &db_song = import_song._db_song;
  db_song[SFI_filename] = DataCell(new_relpath);
  db_song[SFI_volume_id] = DataCell(_volume.get_volume_id());
  db_song[SFI_loaded_time] = DataCellTimestamp(now);
  db_song.record_db();

  if (_canceled || !advance()) {
    // If that was the last song, close the window.
    delete_self();
  }
}

void ImportingDialog::
import_failure() {
  if (!_canceled) {
    new InfoMessage("Failure while importing.");
  }
  delete_self();
}

void ImportingDialog::
cancel_button_clicked() {
  _canceled = true;
  _shell.kill();
}

gint ImportingDialog::
delete_timeout() {
  delete_self();
  return false;
}

void ImportingDialog::
layout_window() {
  set_default_size(200, 0);
  set_title("Importing song files");

  // The box for the main set of widgets.
  vbox_.pack_start(_box1);
  _box1.show();
  _box1.set_border_width(8);

  _box1.pack_start(_progress_bar, true, false, 0);
  _progress_bar.show();
  _box1.pack_start(_song_name, true, false, 0);
  _song_name.show();
  _box1.pack_start(_song_filename, true, false, 0);
  _song_filename.show();

  // The Cancel button.
  _cancel_button.set_usize(80, 36);
  action_area_.pack_start(_cancel_button, true, false, 0);
  _cancel_button.set_flags(GTK_CAN_DEFAULT);
  _cancel_button.show();

  connect_to_method(_cancel_button.clicked, this, &cancel_button_clicked);
}


bool ImportingDialog::
advance() {
  _current_index++;

  return import_current();
}

bool ImportingDialog::
import_current() {
  if (_current_index >= _songs.size()) {
    // All done.
    _current_index = _songs.size();
    _progress_bar.set_value(_current_index);
    _song_name.set_text(string());
    _song_filename.set_text(string());
    return false;
  }

  Song &import_song = *_songs[_current_index];
  TableDataProxy &db_song = import_song._db_song;

  int type = (_convert_to_type >= 0) ? _convert_to_type : import_song._type;
  db_song[SFI_file_type] = DataCell(type, type >= 0);

  _basename_to = "im" + format_int(db_song.get_primary_key());

  // Choose a suitable extension.
  QueryResult qr = music_db->sql_command
    ("SELECT extension FROM soundfile_exts WHERE soundfile_id=" +
     DataCell(type) +
     " ORDER BY extension");
  if (qr.get_num_rows() > 0) {
    _basename_to += qr.get_cell(0, 0).get_string();
  }

  string filename_from = _root_prefix + import_song._relpath;
  _filename_to = _volume.prepare_write_file(_basename_to);
  if (_filename_to.empty()) {
    new InfoMessage("Cannot store directly into volume " +
		    _volume.get_description() + ".");
    return false;
  }

  _progress_bar.set_value(_current_index);
  _song_name.set_text(get_formatted_song_name(db_song));
  _song_filename.set_text(get_basename(filename_from));

  string convert_command;

  if (_convert_to_type < 0) {
    convert_command = "cp " + protect_from_shell(filename_from) + " " + 
      protect_from_shell(_filename_to);

  } else {
    // We have to build a real convert command.
    string convert_command_unexpanded;

    // Get the convert command.
    QueryResult qr = music_db->sql_command
      ("SELECT command FROM soundfile_convert"
       " WHERE from_type=" + DataCell(import_song._type) +
       " AND to_type=" + DataCell(_convert_to_type));
    if (qr.get_num_rows() > 0) {
      convert_command_unexpanded = qr.get_cell(0, 0).get_string();
    }

    if (convert_command_unexpanded.empty()) {
      if (_convert_to_type == import_song._type) {
	convert_command_unexpanded = "cp $from $to";

      } else {
	TableDataProxy from_type_sf(TI_soundfile, import_song._type);
	TableDataProxy to_type_sf(TI_soundfile, _convert_to_type);
	new InfoMessage("Don't know how to convert files from\n" +
			from_type_sf[SFFI_description] + " to " +
			to_type_sf[SFFI_description] + ".  Please\n"
			"define a command for this conversion.\n");
	_canceled = true;
	return false;
      }
    }

    VariableDef vars;
    vars["from"] = protect_from_shell(filename_from);
    vars["to"] = protect_from_shell(_filename_to);
    
    convert_command =
      expand_variables(convert_command_unexpanded, vars);
  }

  _shell.run(convert_command);
  return true;
}
