/* sfClient.C

   Copyright (C) 1999 by David Rose.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   To contact the original author of this program, write to:
           David Rose
           224 W. Dryden St Apt. 418
	   Glendale, CA 91202
	   USA

	   drose@lafn.org
   */

#include "system.h"

#include "sfClient.h"
#include "format.h"
#include "extract_words.h"
#include "infoMessage.h"

#include <unistd.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h> 
#include <errno.h>


static const int sfserver_default_port = 8727;

int SFClient::_port = sfserver_default_port;
string SFClient::_hostname = "localhost";
bool SFClient::_bad_hostname = false;
bool SFClient::_no_server = false;


SFClient::
SFClient() {
  _is_alive = false;
  _can_rstart = true;
  _got_filename = false;
  _is_ready = false;
  _is_playing = false;

  if (_bad_hostname) {
    return;
  }

  // open an Internet stream socket
  _server_socket = socket(AF_INET, SOCK_STREAM, 0);
  if (_server_socket < 0) {
    perror("socket");

  } else {
    struct hostent *server = gethostbyname(_hostname.c_str());
    if (server == NULL) {
      _bad_hostname = true;
      new InfoMessage("Unknown sfserver host: " + _hostname + ".\n"
		      "Paloma will not be able to play any music,\n"
		      "but will work properly otherwise.\n");

    } else {
      // fill in socket structure first
      struct sockaddr_in serv_addr;
      memset(&serv_addr, 0, sizeof(serv_addr));
      memcpy(&serv_addr.sin_addr, server->h_addr, server->h_length);
      serv_addr.sin_family      = server->h_addrtype;
      serv_addr.sin_port        = htons(_port);
      
      // now connect to the server
      int connect_result =
	connect(_server_socket, (struct sockaddr *)&serv_addr,
		sizeof(serv_addr));
      if (connect_result < 0) {
	if (!_no_server) {
	  new InfoMessage
	    ("Unable to contact sfserver at " + _hostname + " port " +
	     format_int(_port) + ".\n"
	     "Paloma will not be able to play any music,\n"
	     "but will work properly otherwise.\n"
	     "Restart sfserver and try again.");
	  _no_server = true;
	}
	close(_server_socket);
	_server_socket = -1;
	
      } else {
	_no_server = false;
	_is_alive = true;
	connect_to_method(Gtk_Main::idle(), this, &get_commands);
      }
    }
  }
}

SFClient::
~SFClient() {
  if (_is_alive) {
    close(_server_socket);
  }
}

bool SFClient::
is_alive() const {
  return _is_alive;
}

void SFClient::
reread_config() {
  send_command("config");
}

void SFClient::
clear() {
  _is_ready = false;
  _got_filename = false;
}

void SFClient::
set_filename(const string &filename, int type, double length) {
  _is_ready = false;
  _got_filename = true;
  send_command("file " + filename + " " + format_int(type) + " " +
	       format_float(length));
}

void SFClient::
set_cd_track(const string &cdrom_device, int track_num, double length) {
  _is_ready = false;
  _got_filename = true;
  send_command("cdrom " + cdrom_device + " " + format_int(track_num) + 
	       " " + format_float(length));
	       
}

bool SFClient::
can_rstart() const {
  return _can_rstart;
}

bool SFClient::
is_ready() const {
  return _is_ready;
}

bool SFClient::
is_playing() const {
  return _is_playing;
}

void SFClient::
play(double start) {
  if (_got_filename) {
    send_command("play " + format_float(start));
  }
}

void SFClient::
play(double start, double stop) {
  if (_got_filename) {
    send_command("play " + format_float(start) + " " + format_float(stop));
  }
}

void SFClient::
stop() {
  send_command("stop");
}

void SFClient::
set_rstart(bool flag) {
  if (flag != _can_rstart) {
    _can_rstart = flag;
    rstart_state(flag);
  }
}

void SFClient::
set_playing(bool flag) {
  if (flag != _is_playing) {
    _is_playing = flag;
    if (flag) {
      start_playing();
    } else {
      stop_playing();
    }
  }
}

void SFClient::
send_command(string command) {
  if (_is_alive) {
    //    cerr << "> " << command << "\n";
    command += "\n";
    write(_server_socket, command.data(), command.length());
  }
}

int SFClient::
get_commands() {
  if (!_is_alive) {
    return true;
  }

  // First, perform a select to see if there is any data waiting for
  // us.

  fd_set fds;
  FD_ZERO(&fds);
  FD_SET(_server_socket, &fds);

  // We must let the select wait a small amount--a half-second,
  // here--just so we don't eat up all the CPU in a spin-wait.

  // It would be better if we could somehow add this socket to the
  // queue that GTK is selecting on.  There doesn't yet seem to be
  // such an interface to GTK, though.

  struct timeval short_instant;
  short_instant.tv_sec = 0;
  short_instant.tv_usec = 500;
  
  int result = select(_server_socket + 1, &fds, 
		      (fd_set *)NULL, (fd_set *)NULL, &short_instant);
  if (result < 0) {
    if (errno != EINTR) {
      perror("select");
    }
    return true;
  }

  if (result != 0) {
    // Ok, there is.  Read it.
    char buffer[4096];
    int bytes_read = read(_server_socket, buffer, 4096);
    if (bytes_read > 0) {
      line_up_server_response(string(buffer, bytes_read));
    }
  }

  return true;
}


// Accept a bunch of output from the server, which might be cut in
// arbitrary places; reassemble it into lines and pass it line-by-line
// into server_response().
void SFClient::
line_up_server_response(const string &input) {
  _last_server_input += input;

  int sol = 0;
  int eol = _last_server_input.find_first_of("\n\r", sol);
  while (eol != string::npos) {
    server_response(_last_server_input.substr(sol, eol - sol));
    sol = eol + 1;
    eol = _last_server_input.find_first_of("\n\r", sol);
  }

  _last_server_input = _last_server_input.substr(sol);
}

void SFClient::
server_response(const string &line) {
  //  cerr << line << "\n";
  string word = extract_first_word(line);

  if (word == "time") {
    update_time(atof(line.substr(4).c_str()));

  } else if (word == "interrupt") {
    set_playing(false);
    interrupt();

  } else if (word == "invalid") {
    set_playing(false);
    _is_ready = false;

  } else if (word == "play") {
    set_playing(true);

  } else if (word == "stop") {
    set_playing(false);

  } else if (word == "end") {
    set_playing(false);
    end();

  } else if (word == "ready") {
    set_rstart(line.find("rstart") != string::npos);
    _is_ready = true;
    set_playing(false);
    ready();
    
  } else if (word == "ERROR") {
    cerr << line << "\n";
   
  } else if (word == "sfserver") {
   
  } else if (word == "config") {
  
  }
}

